class mersenne
{
  unsigned value;
public:
  inline mersenne() { }

  inline mersenne(unsigned u)
  {
    value = u;
  }
  inline mersenne(int i)
  {
    value = i<0 ? i-1 : i;
  }
  inline mersenne(const mersenne &v)
  {
    value = v.value;
  }
  inline
  operator unsigned() const {
    return (value == -1U) ? 0 : value;
  }
  inline
  operator int() const {
    int result;
    return (result = unsigned(*this)) < 0
      ? result+1 : result;
  }
  inline mersenne &
  operator ++()
  {
    if (++value == 0)
      ++value;
    return *this;
  }

  inline mersenne
  operator ++(int)
  {
    unsigned old=value;
    ++*this;
    return old;
  }

  inline mersenne &
  operator --()
  {
    if (value-- == 0)
      --value;
    return *this;
  }

  inline mersenne
  operator --(int)
  {
    unsigned old=value;    
    --*this;
    return old;
  }
  inline mersenne&
  operator *=(const mersenne &b)
  {
    unsigned dummy;
    asm("mull %3;addl %1,%0;adcl %4,%0"
        : "=d,a" (value), "=a,d" (dummy)
        : "%a" (value), "rm" (b.value), "ir" (0)
        : "cc");
    return *this;
  }

  inline friend mersenne
  operator *(const mersenne &a, const mersenne &b)
  {
    unsigned r, dummy;
    asm("mull %3;addl %1,%0;adcl %4,%0"
        : "=d,a" (r), "=a,d" (dummy)
        : "%a" (a.value), "rm" (b.value), "ir" (0)
        : "cc");
    return r;
  }
  inline mersenne &
  operator +=(const mersenne &b)
  {
    asm("addl %2,%0;adcl %3,%0"
        : "=r" (value)
        : "%0" (value), "g" (b.value), "ir" (0)
        : "cc");
    return *this;
  }

  inline friend mersenne
  operator +(const mersenne&a, const mersenne&b)
  {
    unsigned r;
    asm("addl %2,%0;adcl %3,%0"
        : "=r" (r)
        : "%0" (value), "g" (b.value), "ir" (0)
        : "cc");
    return r;
  }
  inline mersenne &
  operator -=(const mersenne &b)
  {
    asm("subl %2,%0;sbbl %3,%0"
        : "=r" (value)
        : "0" (value), "g" (b.value), "ir" (0)
        : "cc");
    return *this;
  }

  inline friend mersenne
  operator -(const mersenne &a)
  {
    return ~a.value;
  }

  inline friend mersenne
  operator -(const mersenne&a, const mersenne&b)
  {
    unsigned r;
    asm("subl %2,%0;sbbl %3,%0"
        : "=r" (r)
        : "0" (value), "g" (b.value), "ir" (0)
        : "cc");
    return r;
  }
  inline friend mersenne
  inverse(const mersenne& b)
  {
    unsigned u1 = 0, u2 = -1, v1=1, v2=b;
    
    for (;;) {
      if (v2 == 0)
        return u1-1;
      u1 -= (u2/v2)*v1;
      u2 %= v2;
      if (u2 == 0)
        return v1;
      v1 -= (v2/u2)*u1;
      v2 %= u2;
    }
  }

  inline friend mersenne
  operator /(const mersenne &a, const mersenne &b)
  {
    return a*inverse(b);
  }
  inline
  mersenne & operator >>=(unsigned width)
  {
    asm("ror %2,%0"
        : "=g" (value)
        : "0" (value), "ic" ((unsigned char)width)
        : "cc");
    return *this;
  }

  inline friend
  mersenne operator >>(unsigned width)
  {
    unsigned r;
    asm("ror %2,%0"
        : "=g" (r)
        : "0" (value), "ic" ((unsigned char)width)
        : "cc");
    return r;
  }

  inline
  mersenne & operator <<=(unsigned width)
  {
    asm("rol %2,%0"
        : "=g" (value)
        : "0" (value), "ic" ((unsigned char)width)
        : "cc");
    return *this;
  }

  inline friend
  mersenne operator <<(unsigned width)
  {
    unsigned r;
    asm("rol %2,%0"
        : "=g" (r)
        : "0" (value), "ic" ((unsigned char)width)
        : "cc");
    return r;
  }
  inline friend mersenne
  operator ^(mersenne b, unsigned e)
  {
    mersenne r = (e&1) ? b : mersenne(1);

    for (;;) {
      if ((e /= 2) == 0)
        return r;
      b *= b;
      if (e&1)
        r *= b;
    }
  }

  inline mersenne &
  operator ^=(unsigned e)
  {
    return (*this = *this^e);
  }
};
